import dataclasses
import re
import warnings
from dataclasses import replace
from typing import Optional, List, Union, Dict, Tuple, Iterable

from starkware.crypto.signature.signature import get_random_private_key
from starkware.starknet.public.abi import get_selector_from_name

from starknet_py.common import create_compiled_contract
from starknet_py.compile.compiler import StarknetCompilationSource
from starknet_py.constants import FEE_CONTRACT_ADDRESS
from starknet_py.net.account.compiled_account_contract import COMPILED_ACCOUNT_CONTRACT
from starknet_py.net.client import Client
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    SentTransactionResponse,
    Hash,
    DeclaredContract,
    Tag,
    TransactionReceipt,
    BlockStateUpdate,
    StarknetBlock,
    Declare,
    Deploy,
    BlockTransactionTraces,
    EstimatedFee,
    Calls,
    TransactionStatus,
    DeployTransactionResponse,
    DeclareTransactionResponse,
    Transaction,
    DeployAccountTransactionResponse,
)
from starknet_py.net.client_utils import _invoke_tx_to_call
from starknet_py.net.gateway_client import GatewayClient
from starknet_py.net.models import (
    InvokeFunction,
    StarknetChainId,
    chain_from_network,
)
from starknet_py.net.models.address import AddressRepresentation, parse_address
from starknet_py.net.models.transaction import DeployAccount
from starknet_py.net.networks import Network, MAINNET, TESTNET
from starknet_py.net.signer import BaseSigner
from starknet_py.net.signer.stark_curve_signer import StarkCurveSigner, KeyPair
from starknet_py.utils.crypto.facade import Call
from starknet_py.utils.data_transformer.execute_transformer import (
    execute_transformer_by_version,
)
from starknet_py.utils.iterable import ensure_iterable
from starknet_py.utils.sync import add_sync_methods
from starknet_py.utils.typed_data import TypedData as TypedDataDataclass
from starknet_py.net.models.typed_data import TypedData


@add_sync_methods
class AccountClient(Client):
    """
    Extends the functionality of :obj:`Client <starknet_py.net.client.Client>`,
    adding additional methods for creating the account contract.
    """

    # pylint: disable=too-many-public-methods
    def __init__(
        self,
        address: AddressRepresentation,
        client: Client,
        signer: Optional[BaseSigner] = None,
        key_pair: Optional[KeyPair] = None,
        chain: Optional[StarknetChainId] = None,
        supported_tx_version: int = 0,
    ):
        """
        :param address: Address of the account contract
        :param client: Instance of Client which will be used to add transactions
        :param signer: Custom signer to be used by AccountClient.
                       If none is provided, default
                       :py:class:`starknet_py.net.signer.stark_curve_signer.StarkCurveSigner` is used.
        :param key_pair: Key pair that will be used to create a default `Signer`
        :param chain: ChainId of the chain used to create the default signer
        :param supported_tx_version: Version of transactions supported by account
        """
        # pylint: disable=too-many-arguments
        if chain is None and signer is None:
            raise ValueError("One of chain or signer must be provided")

        self.address = parse_address(address)
        self.client = client

        if signer is None:
            if key_pair is None:
                raise ValueError(
                    "Either a signer or a key_pair must be provided in AccountClient constructor"
                )

            chain = chain_from_network(net=client.net, chain=chain)
            signer = StarkCurveSigner(
                account_address=self.address, key_pair=key_pair, chain_id=chain
            )
        self.signer = signer
        self.supported_tx_version = supported_tx_version

        if self.supported_tx_version == 0:
            warnings.warn(
                "Account supporting transaction version 0 is deprecated. "
                "Use the new account and set supported_tx_version parameter to 1",
                category=DeprecationWarning,
            )

    @property
    def net(self) -> Network:
        return self.client.net

    async def get_block(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> StarknetBlock:
        return await self.client.get_block(
            block_hash=block_hash, block_number=block_number
        )

    async def get_block_traces(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> BlockTransactionTraces:
        return await self.client.get_block_traces(
            block_hash=block_hash, block_number=block_number
        )

    async def get_state_update(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> BlockStateUpdate:
        return await self.client.get_state_update(
            block_hash=block_hash, block_number=block_number
        )

    async def get_storage_at(
        self,
        contract_address: Hash,
        key: int,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        return await self.client.get_storage_at(
            contract_address=contract_address,
            key=key,
            block_hash=block_hash,
            block_number=block_number,
        )

    async def get_transaction(self, tx_hash: Hash) -> Transaction:
        return await self.client.get_transaction(tx_hash=tx_hash)

    async def get_transaction_receipt(self, tx_hash: Hash) -> TransactionReceipt:
        return await self.client.get_transaction_receipt(tx_hash=tx_hash)

    async def wait_for_tx(
        self,
        tx_hash: Hash,
        wait_for_accept: Optional[bool] = False,
        check_interval=5,
    ) -> Tuple[int, TransactionStatus]:
        return await self.client.wait_for_tx(
            tx_hash=tx_hash,
            wait_for_accept=wait_for_accept,
            check_interval=check_interval,
        )

    async def call_contract(
        self,
        call: Call = None,  # pyright: ignore
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
        *,
        invoke_tx: Call = None,  # pyright: ignore
    ) -> List[int]:
        call = _invoke_tx_to_call(call=call, invoke_tx=invoke_tx)

        return await self.client.call_contract(
            call=call, block_hash=block_hash, block_number=block_number
        )

    async def get_class_hash_at(
        self,
        contract_address: Hash,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        return await self.client.get_class_hash_at(
            contract_address=contract_address,
            block_hash=block_hash,
            block_number=block_number,
        )

    async def get_class_by_hash(self, class_hash: Hash) -> DeclaredContract:
        return await self.client.get_class_by_hash(class_hash=class_hash)

    async def _get_nonce(self) -> int:
        if self.supported_tx_version == 1:
            return await self.get_contract_nonce(self.address, block_hash="latest")

        [nonce] = await self.call_contract(
            Call(
                to_addr=self.address,
                selector=get_selector_from_name("get_nonce"),
                calldata=[],
            ),
            block_hash="pending",
        )
        return nonce

    def _get_default_token_address(self) -> str:
        if self.net not in [TESTNET, MAINNET]:
            raise ValueError(
                "Token_address must be specified when using a custom net address"
            )

        return FEE_CONTRACT_ADDRESS

    async def get_balance(
        self, token_address: Optional[AddressRepresentation] = None
    ) -> int:
        """
        Checks account's balance of specified token.

        :param token_address: Address of the ERC20 contract.
                              If not specified it will be payment token (wrapped ETH) address.
        :return: Token balance
        """

        token_address = token_address or self._get_default_token_address()

        low, high = await self.call_contract(
            Call(
                to_addr=parse_address(token_address),
                selector=get_selector_from_name("balanceOf"),
                calldata=[self.address],
            ),
            block_hash="pending",
        )

        return (high << 128) + low

    async def prepare_invoke_function(
        self,
        calls: Calls,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
        version: Optional[int] = None,
    ) -> InvokeFunction:
        """
        Takes calls and creates InvokeFunction from them

        :param calls: Single call or list of calls
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :param version: Transaction version is supported_tx_version as a default
        :return: InvokeFunction created from the calls (without the signature)

        .. deprecated:: 0.5.0
            This method has been deprecated. Use :meth:`AccountClient.sign_invoke_transaction` to create an already
            signed invoke transactions from calls.
        """
        warnings.warn(
            "prepare_invoke_function has been deprecated. "
            "Use AccountClient.sign_invoke_transaction to create an already signed invoke function.",
            category=DeprecationWarning,
        )

        if version is None:
            version = self.supported_tx_version

        self._assert_version_matches_supported_tx_version(version)

        if version == 0:
            warnings.warn(
                "Transaction with version 0 is deprecated and will be removed in the future. "
                "Use AccountClient supporting the transaction version 1",
                category=DeprecationWarning,
            )

        nonce = await self._get_nonce()

        calldata_py = merge_calls(ensure_iterable(calls))

        if version == 0:
            calldata_py.append(nonce)

        execute_transformer = execute_transformer_by_version(version)
        wrapped_calldata, _ = execute_transformer.from_python(*calldata_py)

        transaction = make_invoke_function_by_version(
            contract_address=self.address,
            calldata=wrapped_calldata,
            signature=[],
            max_fee=0,
            version=version,
            entry_point_selector=get_selector_from_name("__execute__"),
            nonce=None if version == 0 else nonce,
        )

        max_fee = await self._get_max_fee(transaction, max_fee, auto_estimate)

        return dataclasses.replace(transaction, max_fee=max_fee)

    async def _get_max_fee(
        self,
        transaction: Union[InvokeFunction, Declare, DeployAccount],
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> int:
        if auto_estimate and max_fee is not None:
            raise ValueError(
                "Max_fee and auto_estimate are exclusive and cannot be provided at the same time."
            )

        if (
            isinstance(transaction, (Declare, DeployAccount))
            and transaction.version != 1
        ):
            raise ValueError(
                "Estimating fee for Declare/DeployAccount transactions with versions other than 1 is not supported."
            )

        if auto_estimate:
            estimate_fee = await self.estimate_fee(transaction)
            max_fee = int(estimate_fee.overall_fee * 1.1)

        if max_fee is None:
            raise ValueError("Max_fee must be specified when invoking a transaction")

        return max_fee

    async def sign_invoke_transaction(
        self,
        calls: Calls,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
        version: Optional[int] = None,
    ) -> InvokeFunction:
        """
        Takes calls and creates signed InvokeFunction

        :param calls: Single call or list of calls
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :param version: Transaction version
        :return: InvokeFunction created from the calls
        """
        if version is None:
            version = self.supported_tx_version

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            execute_tx = await self.prepare_invoke_function(
                calls, max_fee, auto_estimate, version
            )

        signature = self.signer.sign_transaction(execute_tx)
        execute_tx = add_signature_to_transaction(execute_tx, signature)

        return execute_tx

    async def sign_declare_transaction(
        self,
        compilation_source: Optional[StarknetCompilationSource] = None,
        compiled_contract: Optional[str] = None,
        cairo_path: Optional[List[str]] = None,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> Declare:
        """
        Create and sign declare transaction.

        Either `compilation_source` or `compiled_contract` is required.

        :param compilation_source: string containing source code or a list of source files paths
        :param compiled_contract: string containing compiled contract bytecode.
                                  Useful for reading compiled contract from a file
        :param cairo_path: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :return: Signed Declare transaction
        """
        # pylint: disable=too-many-arguments
        if self.supported_tx_version != 1:
            raise ValueError(
                "Signing declare transactions is only supported with transaction version 1"
            )

        compiled_contract = create_compiled_contract(
            compilation_source, compiled_contract, cairo_path
        )
        declare_tx = Declare(
            contract_class=compiled_contract,
            sender_address=self.address,
            max_fee=0,
            signature=[],
            nonce=await self._get_nonce(),
            version=self.supported_tx_version,
        )

        max_fee = await self._get_max_fee(
            transaction=declare_tx, max_fee=max_fee, auto_estimate=auto_estimate
        )
        declare_tx = dataclasses.replace(declare_tx, max_fee=max_fee)
        signature = self.signer.sign_transaction(declare_tx)

        return dataclasses.replace(declare_tx, signature=signature, max_fee=max_fee)

    async def sign_deploy_account_transaction(
        self,
        *,
        class_hash: int,
        contract_address_salt: int,
        constructor_calldata: Optional[List[int]] = None,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> DeployAccount:
        """
        Create and sign deploy account transaction

        :param class_hash: Class hash of the contract class to be deployed
        :param contract_address_salt: A salt used to calculate deployed contract address
        :param constructor_calldata: Calldata to be passed to contract constructor
            and used to calculate deployed contract address
        :param max_fee: Max fee to be paid for deploying account transaction. Enough tokens must be prefunded before
            sending the transaction for it to succeed.
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :return: Signed DeployAccount transaction
        """
        if self.supported_tx_version != 1:
            raise ValueError(
                "Signing deploy account transactions is only supported with transaction version 1"
            )

        constructor_calldata = constructor_calldata or []

        deploy_account_tx = DeployAccount(
            class_hash=class_hash,
            contract_address_salt=contract_address_salt,
            constructor_calldata=constructor_calldata,
            version=self.supported_tx_version,
            max_fee=0,
            signature=[],
            nonce=0,
        )

        max_fee = await self._get_max_fee(
            transaction=deploy_account_tx, max_fee=max_fee, auto_estimate=auto_estimate
        )
        deploy_account_tx = dataclasses.replace(deploy_account_tx, max_fee=max_fee)
        signature = self.signer.sign_transaction(deploy_account_tx)

        return dataclasses.replace(deploy_account_tx, signature=signature)

    async def send_transaction(
        self, transaction: InvokeFunction
    ) -> SentTransactionResponse:
        if transaction.max_fee == 0:
            warnings.warn(
                "Transaction will fail with max_fee set to 0. Change it to a higher value.",
                DeprecationWarning,
            )

        return await self.client.send_transaction(transaction=transaction)

    async def execute(
        self,
        calls: Calls,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
        version: Optional[int] = None,
    ) -> SentTransactionResponse:
        """
        Takes calls and executes transaction

        :param calls: Single call or list of calls
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :param version: Transaction version
        :return: SentTransactionResponse
        """
        execute_transaction = await self.sign_invoke_transaction(
            calls, max_fee, auto_estimate, version
        )
        return await self.send_transaction(execute_transaction)

    async def deploy(self, transaction: Deploy) -> DeployTransactionResponse:
        return await self.client.deploy(transaction=transaction)

    async def deploy_account(
        self, transaction: DeployAccount
    ) -> DeployAccountTransactionResponse:
        return await self.client.deploy_account(transaction=transaction)

    async def declare(self, transaction: Declare) -> DeclareTransactionResponse:
        return await self.client.declare(transaction=transaction)

    async def estimate_fee(
        self,
        tx: Union[InvokeFunction, Declare, DeployAccount],
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> EstimatedFee:
        """
        :param tx: Transaction which fee we want to calculate
        :param block_hash: Estimate fee at specific block hash
        :param block_number: Estimate fee at given block number
            (or "latest" / "pending" for the latest / pending block), default is "pending"
        :return: Estimated fee
        """
        signature = self.signer.sign_transaction(tx)
        tx = add_signature_to_transaction(tx, signature)

        return await self.client.estimate_fee(
            tx=tx,
            block_hash=block_hash,
            block_number=block_number,
        )

    async def get_code(self, *args, **kwargs):
        warnings.warn(
            "get_code was removed from Client interface and will be removed from AccountClient in future versions",
            category=DeprecationWarning,
        )
        if not isinstance(self.client, GatewayClient):
            raise TypeError("AccountClient.get_code only supports using GatewayClient")

        return await self.client.get_code(*args, **kwargs)

    def _assert_version_matches_supported_tx_version(self, version: int):
        if version != self.supported_tx_version:
            raise ValueError(
                f"Provided version: {version} is not equal to account's "
                f"supported_tx_version: {self.supported_tx_version}"
            )

    @staticmethod
    async def create_account(
        client: Client,
        private_key: Optional[int] = None,
        signer: Optional[BaseSigner] = None,
        chain: Optional[StarknetChainId] = None,
    ) -> "AccountClient":
        """
        Creates the account using
        `OpenZeppelin Account contract
        <https://github.com/starkware-libs/cairo-lang/blob/4e233516f52477ad158bc81a86ec2760471c1b65/src/starkware/starknet/third_party/open_zeppelin/Account.cairo>`_

        :param client: Instance of Client (i.e. FullNodeClient or GatewayClient)
                       which will be used to add the transactions
        :param private_key: Private Key used for the account
        :param signer: Signer used to create account and sign transaction
        :param chain: ChainId of the chain used to create the default signer
        :return: Instance of AccountClient which interacts with created account on given network

        .. deprecated:: 0.5.0
            This method has been deprecated and will be deleted once transaction version 0 is removed.
            Compiled account contract will no longer be bundled with StarkNet.py.
            Consider transitioning to deploying account contract of choice and creating AccountClient
            through a constructor.
        """
        warnings.warn(
            "Account deployment through AccountClient is deprecated and will be deleted once transaction version "
            "0 is removed. Consider transitioning to creating AccountClient through a constructor.",
            category=DeprecationWarning,
        )

        if chain is None and signer is None:
            raise ValueError("One of chain or signer must be provided")

        if signer is None:
            chain = chain_from_network(net=client.net, chain=chain)
            used_private_key = private_key or get_random_private_key()
            key_pair = KeyPair.from_private_key(used_private_key)
            address = await deploy_account_contract(client, key_pair.public_key)
            signer = StarkCurveSigner(
                account_address=address, key_pair=key_pair, chain_id=chain
            )
        else:
            address = await deploy_account_contract(client, signer.public_key)

        version = get_account_version()

        return AccountClient(
            client=client,
            address=address,
            signer=signer,
            supported_tx_version=version,
        )

    async def get_contract_nonce(
        self,
        contract_address: int,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        return await self.client.get_contract_nonce(
            contract_address, block_hash, block_number
        )

    def sign_message(self, typed_data: TypedData) -> List[int]:
        """
        Sign an TypedData TypedDict for off-chain usage with the starknet private key and return the signature
        This adds a message prefix, so it can't be interchanged with transactions

        :param typed_data: TypedData TypedDict to be signed
        :return: The signature of the TypedData TypedDict
        """
        return self.signer.sign_message(typed_data, self.address)

    def hash_message(self, typed_data: TypedData) -> int:
        """
        Hash a TypedData TypedDict with pedersen hash and return the hash
        This adds a message prefix, so it can't be interchanged with transactions

        :param typed_data: TypedData TypedDict to be hashed
        :return: the hash of the TypedData TypedDict
        """
        typed_data_dataclass: TypedDataDataclass = TypedDataDataclass.from_dict(
            typed_data
        )
        return typed_data_dataclass.message_hash(self.address)

    async def verify_message(self, typed_data: TypedData, signature: List[int]) -> bool:
        """
        Verify a signature of a TypedData TypedDict

        :param typed_data: TypedData TypedDict to be verified
        :param signature: signature of the TypedData TypedDict
        :return: true if the signature is valid, false otherwise
        """
        msg_hash = self.hash_message(typed_data)
        return await self._verify_message_hash(msg_hash, signature)

    async def _verify_message_hash(self, msg_hash: int, signature: List[int]) -> bool:
        """
        Verify a signature of a given hash

        :param msg_hash: hash to be verified
        :param signature: signature of the hash
        :return: true if the signature is valid, false otherwise
        """
        calldata = [msg_hash, len(signature), *signature]

        call = Call(
            to_addr=self.address,
            selector=get_selector_from_name("is_valid_signature"),
            calldata=calldata,
        )
        try:
            await self.call_contract(call=call, block_hash="pending")
            return True
        except ClientError as ex:
            if re.search(r"Signature\s.+,\sis\sinvalid", ex.message):
                return False
            raise ex


async def deploy_account_contract(
    client: Client, public_key: int
) -> AddressRepresentation:
    # pylint: disable=import-outside-toplevel
    # FIXME move this import to top once circular import is resolved
    from starknet_py.transactions.deploy import make_deploy_tx

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        deploy_tx = make_deploy_tx(
            constructor_calldata=[public_key],
            compiled_contract=COMPILED_ACCOUNT_CONTRACT,
        )

    result = await client.deploy(deploy_tx)
    await client.wait_for_tx(
        tx_hash=result.transaction_hash,
    )
    return result.contract_address


def add_signature_to_transaction(
    tx: InvokeFunction, signature: List[int]
) -> InvokeFunction:
    return replace(tx, signature=signature)


def merge_calls(calls: Iterable[Call]) -> List:
    def parse_call(
        call: Call, current_data_len: int, entire_calldata: List
    ) -> Tuple[Dict, int, List]:
        data = {
            "to": call.to_addr,
            "selector": call.selector,
            "data_offset": current_data_len,
            "data_len": len(call.calldata),
        }
        current_data_len += len(call.calldata)
        entire_calldata += call.calldata

        return data, current_data_len, entire_calldata

    calldata = []
    current_data_len = 0
    entire_calldata = []
    for call in calls:
        data, current_data_len, entire_calldata = parse_call(
            call, current_data_len, entire_calldata
        )
        calldata.append(data)

    return [calldata, entire_calldata]


def make_invoke_function_by_version(
    # pylint: disable=too-many-arguments
    contract_address: AddressRepresentation,
    calldata: List[int],
    signature: List[int],
    max_fee: int,
    version: int,
    nonce: Optional[int],
    entry_point_selector: int,
) -> InvokeFunction:
    params = {
        "calldata": calldata,
        "signature": signature,
        "max_fee": max_fee,
        "version": version,
        "nonce": nonce,
        "contract_address": contract_address,
    }

    if version == 0:
        params["entry_point_selector"] = entry_point_selector

    return InvokeFunction(**params)


def get_account_version():
    return 1 if "__validate__" in COMPILED_ACCOUNT_CONTRACT else 0



from typing import List, Union, Optional

from starkware.starknet.public.abi_structs import identifier_manager_from_abi

from starknet_py.utils.data_transformer import FunctionCallSerializer


def translate_constructor_args(
    abi: List, constructor_args: Optional[Union[List, dict]]
) -> List[int]:
    constructor_abi = next(
        (member for member in abi if member["type"] == "constructor"),
        None,
    )

    # Constructor might not accept any arguments
    if not constructor_abi or not constructor_abi["inputs"]:
        return []

    if not constructor_args:
        raise ValueError(
            "Provided contract has a constructor and no arguments were provided."
        )

    args, kwargs = (
        ([], constructor_args)
        if isinstance(constructor_args, dict)
        else (constructor_args, {})
    )
    calldata, _args = FunctionCallSerializer(
        constructor_abi, identifier_manager_from_abi(abi)
    ).from_python(*args, **kwargs)
    return calldata



from starkware.starknet.services.api.feeder_gateway.response_objects import (
    TransactionStatus,
)

TxStatus = TransactionStatus

ACCEPTED_STATUSES = (TxStatus.ACCEPTED_ON_L1, TxStatus.ACCEPTED_ON_L2)

# Address came from starkware-libs/starknet-addresses repository: https://github.com/starkware-libs/starknet-addresses
FEE_CONTRACT_ADDRESS = (
    "0x049d36570d4e46f48e99674bd3fcc84644ddd6b96f7c741b1562b82f9e004dc7"
)

DEFAULT_DEPLOYER_ADDRESS = (
    "0x041a78e741e5aF2fEc34B695679bC6891742439f7AFB8484Ecd7766661aD02BF"
)

RPC_INVALID_MESSAGE_SELECTOR_ERROR = 21
RPC_CLASS_HASH_NOT_FOUND_ERROR = 28



import warnings
from typing import Union, Optional
from typing_extensions import get_args

from starknet_py.net.client_models import Hash, Tag, Call


def hash_to_felt(value: Hash) -> str:
    """
    Convert hash to hexadecimal string
    """
    if isinstance(value, str):
        return value

    return hex(value)


def is_block_identifier(value: Union[int, Hash, Tag]) -> bool:
    return isinstance(value, str) and value in get_args(Tag)


def _invoke_tx_to_call(
    call: Optional[Call] = None, invoke_tx: Optional[Call] = None
) -> Call:
    if call is not None and invoke_tx is not None:
        raise ValueError("invoke_tx and call are mutually exclusive")

    if invoke_tx is not None:
        warnings.warn(
            "invoke_tx parameter is deprecated. Use call instead",
            category=DeprecationWarning,
        )
        call = invoke_tx

    if call is None:
        raise ValueError("Call is a required parameter")

    return call



from typing import Optional, List, Union

from starkware.starknet.services.api.contract_class import ContractClass

from starknet_py.compile.compiler import (
    StarknetCompilationSource,
    Compiler,
    create_contract_class,
)


def create_compiled_contract(
    compilation_source: Optional[StarknetCompilationSource] = None,
    compiled_contract: Optional[str] = None,
    search_paths: Optional[List[str]] = None,
) -> ContractClass:
    if not compiled_contract:
        if not compilation_source:
            raise ValueError(
                "One of compiled_contract or compilation_source is required."
            )

        compiled_contract = Compiler(
            contract_source=compilation_source, cairo_path=search_paths
        ).compile_contract()
    definition = create_contract_class(compiled_contract)
    return definition


def int_from_hex(number: Union[str, int]) -> int:
    return number if isinstance(number, int) else int(number, 16)



from typing import Union, Sequence

from starkware.starknet.core.os.contract_address.contract_address import (
    calculate_contract_address_from_hash,
)
from starkware.starknet.services.api.feeder_gateway.feeder_gateway_client import (
    BlockIdentifier,
)


from starknet_py.utils.docs import as_our_module

AddressRepresentation = Union[int, str]
Address = int


def parse_address(value: AddressRepresentation) -> Address:
    if isinstance(value, int):
        return value

    try:
        return int(value, 16)
    except TypeError as t_err:
        raise TypeError("Invalid address format.") from t_err


def compute_address(
    *,
    class_hash: int,
    constructor_calldata: Sequence[int],
    salt: int,
    deployer_address: int = 0,
) -> int:
    """
    Computes contract's address.

    :param class_hash: class hash of the contract
    :param constructor_calldata: calldata for the contract constructor
    :param salt: salt used to calculate contract address
    :param deployer_address: address of the deployer (if not provided default 0 is used)
    :return: Contract's address
    """

    return calculate_contract_address_from_hash(
        class_hash=class_hash,
        constructor_calldata=constructor_calldata,
        salt=salt,
        deployer_address=deployer_address,
    )


BlockIdentifier = as_our_module(BlockIdentifier)



from typing import TypedDict, Union, Literal

MAINNET = "mainnet"
TESTNET = "testnet"

PredefinedNetwork = Literal["mainnet", "testnet"]


class CustomGatewayUrls(TypedDict):
    feeder_gateway_url: str
    gateway_url: str


Network = Union[PredefinedNetwork, str, CustomGatewayUrls]


def net_address_from_net(net: str) -> str:
    return {
        MAINNET: "https://alpha-mainnet.starknet.io",
        TESTNET: "https://alpha4.starknet.io",
    }.get(net, net)



import warnings
from enum import Enum
from typing import List, TypedDict, Tuple

from starknet_py.constants import RPC_CLASS_HASH_NOT_FOUND_ERROR
from starknet_py.net.client import Client
from starknet_py.net.client_errors import ContractNotFoundError, ClientError
from starknet_py.net.client_models import Abi, DeclaredContract
from starknet_py.net.models import Address
from starknet_py.proxy.proxy_check import (
    OpenZeppelinProxyCheck,
    ArgentProxyCheck,
    ProxyCheck,
)


class ProxyConfig(TypedDict, total=False):
    """
    Proxy resolving configuration
    """

    proxy_checks: List[ProxyCheck]
    """
    List of classes implementing :class:`starknet_py.proxy.proxy_check.ProxyCheck` ABC,
    that will be used for checking if contract at the address is a proxy contract.
    """


def prepare_proxy_config(proxy_config: ProxyConfig) -> ProxyConfig:
    if "max_steps" in proxy_config:
        warnings.warn(
            "ProxyConfig.max_steps is deprecated. Contract.from_address always makes at most 1 step.",
            category=DeprecationWarning,
        )
    if "proxy_checks" in proxy_config:
        return proxy_config

    proxy_checks = [OpenZeppelinProxyCheck(), ArgentProxyCheck()]
    return {"proxy_checks": proxy_checks}


class ImplementationType(Enum):
    """
    Enum representing transaction types
    """

    CLASS_HASH = "class_hash"
    ADDRESS = "address"


class ContractAbiResolver:
    """
    Class for resolving abi of a contract
    """

    def __init__(
        self,
        address: Address,
        client: Client,
        proxy_config: ProxyConfig,
    ):
        """
        :param address: Contract's address
        :param client: Client used for resolving abi
        :param proxy_config: Proxy config for resolving proxy
        """
        self.address = address
        self.client = client
        self.proxy_config = proxy_config

    async def resolve(self) -> Abi:
        """
        Returns abi of either direct contract or contract proxied by direct contract depending on proxy_config.
        :raises ContractNotFoundError: when contract could not be found at address
        :raises ProxyResolutionError: when given ProxyChecks were not sufficient to resolve proxy
        :raises AbiNotFoundError: when abi is not present in contract class at address
        """
        if len(self.proxy_config) == 0:
            return await self.get_abi_for_address()
        return await self.resolve_abi()

    async def get_abi_for_address(self) -> Abi:
        """
        Returns abi of a contract directly from address.
        :raises ContractNotFoundError: when contract could not be found at address
        :raises AbiNotFoundError: when abi is not present in contract class at address
        """
        contract_class = await self._get_class_by_address(address=self.address)
        if contract_class.abi is None:
            raise AbiNotFoundError()
        return contract_class.abi

    async def resolve_abi(self) -> Abi:
        """
        Returns abi of a contract that is being proxied by contract at address.
        :raises ContractNotFoundError: when contract could not be found at address
        :raises ProxyResolutionError: when given ProxyChecks were not sufficient to resolve proxy
        :raises AbiNotFoundError: when abi is not present in proxied contract class at address
        """
        impl = await self._get_implementation_from_proxy()
        implementation, implementation_type = impl

        if implementation_type == ImplementationType.CLASS_HASH:
            contract_class = await self.client.get_class_by_hash(implementation)
        else:
            contract_class = await self._get_class_by_address(address=implementation)

        if contract_class.abi is None:
            raise AbiNotFoundError()
        return contract_class.abi

    async def _get_implementation_from_proxy(self) -> Tuple[int, ImplementationType]:
        proxy_checks = self.proxy_config.get("proxy_checks", [])
        for proxy_check in proxy_checks:
            implementation = await proxy_check.implementation_hash(
                address=self.address, client=self.client
            )
            if implementation is not None:
                return implementation, ImplementationType.CLASS_HASH

            implementation = await proxy_check.implementation_address(
                address=self.address, client=self.client
            )
            if implementation is not None:
                return implementation, ImplementationType.ADDRESS

        raise ProxyResolutionError()

    async def _get_class_by_address(self, address: Address) -> DeclaredContract:
        try:
            contract_class_hash = await self.client.get_class_hash_at(
                contract_address=address
            )
            contract_class = await self.client.get_class_by_hash(
                class_hash=contract_class_hash
            )
        except ClientError as err:
            if (
                "is not deployed" in err.message
                or err.code == RPC_CLASS_HASH_NOT_FOUND_ERROR
            ):
                raise ContractNotFoundError(address=address) from err
            raise err

        return contract_class


class AbiNotFoundError(Exception):
    """
    Error while resolving contract abi.
    """


class ProxyResolutionError(Exception):
    """
    Error while resolving proxy using ProxyChecks.
    """

    def __init__(self):
        self.message = "Couldn't resolve proxy using given ProxyChecks"
        super().__init__(self.message)



from __future__ import annotations

import dataclasses
import warnings
from dataclasses import dataclass
from typing import (
    List,
    Optional,
    TypeVar,
    Union,
    Dict,
    NamedTuple,
)

from starkware.cairo.lang.compiler.identifier_manager import IdentifierManager
from starkware.starknet.core.os.class_hash import compute_class_hash
from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.public.abi_structs import identifier_manager_from_abi
from starkware.starknet.services.api.feeder_gateway.feeder_gateway_client import (
    CastableToHash,
)

from starknet_py.common import create_compiled_contract
from starknet_py.compile.compiler import StarknetCompilationSource
from starknet_py.constants import DEFAULT_DEPLOYER_ADDRESS
from starknet_py.net.udc_deployer.deployer import Deployer
from starknet_py.proxy.contract_abi_resolver import (
    ProxyConfig,
    ContractAbiResolver,
    prepare_proxy_config,
)
from starknet_py.net import AccountClient
from starknet_py.net.client import Client
from starknet_py.net.client_models import Hash, Tag
from starknet_py.net.models import (
    InvokeFunction,
    AddressRepresentation,
    parse_address,
    compute_address,
)
from starknet_py.transactions.deploy import make_deploy_tx
from starknet_py.utils.contructor_args_translator import translate_constructor_args
from starknet_py.utils.crypto.facade import pedersen_hash, Call
from starknet_py.utils.data_transformer import FunctionCallSerializer
from starknet_py.utils.sync import add_sync_methods

ABI = list
ABIEntry = dict
TSentTransaction = TypeVar("TSentTransaction", bound="SentTransaction")


@dataclass(frozen=True)
class ContractData:
    address: int
    abi: ABI
    identifier_manager: IdentifierManager

    @staticmethod
    def from_abi(address: int, abi: ABI) -> "ContractData":
        return ContractData(
            address=address,
            abi=abi,
            identifier_manager=identifier_manager_from_abi(abi),
        )


@add_sync_methods
@dataclass(frozen=True)
class SentTransaction:
    """
    Dataclass exposing the interface of transaction related to a performed action
    """

    hash: CastableToHash
    _client: Client
    status: Optional[str] = None
    block_number: Optional[int] = None

    async def wait_for_acceptance(
        self: TSentTransaction,
        wait_for_accept: Optional[bool] = False,
        check_interval=5,
    ) -> TSentTransaction:
        """
        Waits for transaction to be accepted on chain. By default, returns when status is ``PENDING`` -
        use ``wait_for_accept`` to wait till ``ACCEPTED`` status.
        Returns a new SentTransaction instance, **does not mutate original instance**.
        """
        block_number, status = await self._client.wait_for_tx(
            self.hash,
            wait_for_accept=wait_for_accept,
            check_interval=check_interval,
        )
        return dataclasses.replace(
            self,
            status=status,
            block_number=block_number,
        )


@add_sync_methods
@dataclass(frozen=True)
class InvokeResult(SentTransaction):
    # We ensure these are not None in __post_init__
    contract: ContractData = None  # pyright: ignore
    invoke_transaction: InvokeFunction = None  # pyright: ignore

    def __post_init__(self):
        assert self.contract is not None
        assert self.invoke_transaction is not None


InvocationResult = InvokeResult


@add_sync_methods
@dataclass(frozen=True)
class DeclareResult(SentTransaction):
    _account: AccountClient = None  # pyright: ignore
    class_hash: int = None  # pyright: ignore
    compiled_contract: str = None  # pyright: ignore

    def __post_init__(self):
        if any(
            field is None
            for field in [self.class_hash, self._account, self.compiled_contract]
        ):
            raise ValueError(
                "None of the account, class_hash and compiled_contract fields can be None"
            )

    async def deploy(
        self,
        *,
        deployer_address: AddressRepresentation = DEFAULT_DEPLOYER_ADDRESS,
        salt: Optional[int] = None,
        unique: bool = True,
        constructor_args: Optional[Union[List, Dict]] = None,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> "DeployResult":
        """
        Deploys a contract

        :param deployer_address: Address of the UDC. Is set to the address of
            the default UDC (same address on mainnet/testnet/devnet) by default.
            Must be set when using custom network other than ones listed above.
        :param salt: Optional salt. Random value is selected if it is not provided.
        :param unique: Determines if the contract should be salted with the account address.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param max_fee: Max amount of Wei to be paid when executing transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :return: DeployResult instance
        """
        # pylint: disable=too-many-arguments
        abi = create_compiled_contract(compiled_contract=self.compiled_contract).abi

        deployer = Deployer(
            deployer_address=deployer_address,
            account_address=self._account.address if unique else None,
        )
        deploy_call, address = deployer.create_deployment_call(
            class_hash=self.class_hash, salt=salt, abi=abi, calldata=constructor_args
        )
        res = await self._account.execute(
            calls=deploy_call, max_fee=max_fee, auto_estimate=auto_estimate
        )

        deployed_contract = Contract(
            client=self._account,
            address=address,
            abi=abi,
        )
        deploy_result = DeployResult(
            hash=res.transaction_hash,
            _client=self._account.client,
            deployed_contract=deployed_contract,
        )

        return deploy_result


@add_sync_methods
@dataclass(frozen=True)
class DeployResult(SentTransaction):
    # We ensure this is not None in __post_init__
    deployed_contract: "Contract" = None  # pyright: ignore

    def __post_init__(self):
        if self.deployed_contract is None:
            raise ValueError("deployed_contract can't be None")


# pylint: disable=too-many-instance-attributes
@add_sync_methods
class PreparedFunctionCall(Call):
    def __init__(
        self,
        calldata: List[int],
        arguments: Dict[str, List[int]],
        selector: int,
        client: Client,
        payload_transformer: FunctionCallSerializer,
        contract_data: ContractData,
        max_fee: Optional[int],
        version: int,
    ):
        # pylint: disable=too-many-arguments
        super().__init__(
            to_addr=contract_data.address, selector=selector, calldata=calldata
        )
        self.arguments = arguments
        self._client = client
        self._payload_transformer = payload_transformer
        self._contract_data = contract_data
        self.max_fee = max_fee
        self.version = version

    async def call_raw(
        self,
        block_hash: Optional[str] = None,
    ) -> List[int]:
        """
        Calls a method without translating the result into python values.

        :param block_hash: Optional block hash
        :return: list of ints
        """
        return await self._client.call_contract(call=self, block_hash=block_hash)

    async def call(
        self,
        block_hash: Optional[str] = None,
    ) -> NamedTuple:
        """
        Calls a method.

        :param block_hash: Optional block hash
        :return: CallResult or List[int] if return_raw is used
        """
        result = await self.call_raw(block_hash=block_hash)
        return self._payload_transformer.to_python(result)

    async def invoke(
        self,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> InvokeResult:
        """
        Invokes a method.

        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        :return: InvokeResult
        """
        if max_fee is not None:
            self.max_fee = max_fee

        transaction = await self._account_client.sign_invoke_transaction(
            calls=self,
            max_fee=self.max_fee,
            auto_estimate=auto_estimate,
            version=self.version,
        )
        response = await self._account_client.send_transaction(transaction)

        invoke_result = InvokeResult(
            hash=response.transaction_hash,  # noinspection PyTypeChecker
            _client=self._account_client,
            contract=self._contract_data,
            invoke_transaction=transaction,
        )

        return invoke_result

    async def estimate_fee(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ):
        """
        Estimate fee for prepared function call

        :param block_hash: Estimate fee at specific block hash
        :param block_number: Estimate fee at given block number
            (or "latest" / "pending" for the latest / pending block), default is "pending"
        :return: Estimated amount of Wei executing specified transaction will cost
        :raises ValueError: when max_fee of PreparedFunctionCall is not None or 0.
        """
        if self.max_fee is not None and self.max_fee != 0:
            raise ValueError(
                "Cannot estimate fee of PreparedFunctionCall with max_fee not None or 0."
            )

        tx = await self._account_client.sign_invoke_transaction(
            calls=self, max_fee=0, version=self.version
        )

        return await self._account_client.estimate_fee(
            tx=tx, block_hash=block_hash, block_number=block_number
        )

    @property
    def _account_client(self) -> AccountClient:
        client = self._client
        if isinstance(client, AccountClient):
            return client

        raise ValueError(
            "Contract uses an account that can't invoke transactions. You need to use AccountClient for that."
        )


@add_sync_methods
class ContractFunction:
    def __init__(
        self, name: str, abi: ABIEntry, contract_data: ContractData, client: Client
    ):
        self.name = name
        self.abi = abi
        self.inputs = abi["inputs"]
        self.contract_data = contract_data
        self._client = client
        self._payload_transformer = FunctionCallSerializer(
            abi=self.abi, identifier_manager=self.contract_data.identifier_manager
        )

    def prepare(
        self,
        *args,
        version: Optional[int] = None,
        max_fee: Optional[int] = None,
        **kwargs,
    ) -> PreparedFunctionCall:
        """
        ``*args`` and ``**kwargs`` are translated into Cairo calldata.
         Creates a ``PreparedFunctionCall`` instance
         which exposes calldata for every argument and adds more arguments when calling methods.

        :param version: PreparedFunctionCall version
        :param max_fee: Max amount of Wei to be paid when executing transaction
        :return: PreparedFunctionCall
        """
        if version is None:
            version = (
                self._client.supported_tx_version
                if isinstance(self._client, AccountClient)
                else 0
            )

        if version == 0:
            warnings.warn(
                "Transaction with version 0 is deprecated and will be removed in the future. "
                "Use AccountClient supporting the transaction version 1",
                category=DeprecationWarning,
            )

        calldata, arguments = self._payload_transformer.from_python(*args, **kwargs)
        return PreparedFunctionCall(
            calldata=calldata,
            arguments=arguments,
            contract_data=self.contract_data,
            client=self._client,
            payload_transformer=self._payload_transformer,
            selector=self.get_selector(self.name),
            max_fee=max_fee,
            version=version,
        )

    async def call(
        self,
        *args,
        block_hash: Optional[str] = None,
        version: Optional[int] = None,
        **kwargs,
    ) -> NamedTuple:
        """
        :param block_hash: Block hash to execute the contract at specific point of time
        :param version: Call version

        Call contract's function. ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        The result is translated from Cairo data to python values.
        Equivalent of ``.prepare(*args, **kwargs).call()``.
        """
        return await self.prepare(max_fee=0, version=version, *args, **kwargs).call(
            block_hash=block_hash
        )

    async def invoke(
        self,
        *args,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
        **kwargs,
    ) -> InvokeResult:
        """
        Invoke contract's function. ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        Equivalent of ``.prepare(*args, **kwargs).invoke()``.

        :param max_fee: Max amount of Wei to be paid when executing transaction
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs
        """
        prepared_call = self.prepare(*args, **kwargs)
        return await prepared_call.invoke(max_fee=max_fee, auto_estimate=auto_estimate)

    @staticmethod
    def get_selector(function_name: str):
        """
        :param function_name: Contract function's name
        :return: A StarkNet integer selector for this function inside the contract
        """
        return get_selector_from_name(function_name)


FunctionsRepository = Dict[str, ContractFunction]


@add_sync_methods
class Contract:
    """
    Cairo contract's model.
    """

    def __init__(self, address: AddressRepresentation, abi: list, client: Client):
        """
        Should be used instead of ``from_address`` when ABI is known statically.

        :param address: contract's address
        :param abi: contract's abi
        :param client: client used for API calls
        """
        self.data = ContractData.from_abi(parse_address(address), abi)
        self._functions = self._make_functions(self.data, client)
        self.client = client

    @property
    def functions(self) -> FunctionsRepository:
        """
        :return: All functions exposed from a contract.
        """
        return self._functions

    @property
    def address(self) -> int:
        return self.data.address

    @staticmethod
    async def from_address(
        address: AddressRepresentation,
        client: Client,
        proxy_config: Union[bool, ProxyConfig] = False,
    ) -> "Contract":
        """
        Fetches ABI for given contract and creates a new Contract instance with it. If you know ABI statically you
        should create Contract's instances directly instead of using this function to avoid unnecessary API calls.

        :raises ContractNotFoundError: when contract is not found
        :raises TypeError: when given client's `get_class_by_hash` method does not return abi
        :raises ProxyResolutionError: when given ProxyChecks were not sufficient to resolve proxy's implementation
        :param address: Contract's address
        :param client: Client
        :param proxy_config: Proxy resolving config
            If set to ``True``, will use default proxy checks
            :class:`starknet_py.proxy.proxy_check.OpenZeppelinProxyCheck`
            and :class:`starknet_py.proxy.proxy_check.ArgentProxyCheck`.

            If set to ``False``, :meth:`Contract.from_address` will not resolve proxies.

            If a valid :class:`starknet_py.contract_abi_resolver.ProxyConfig` is provided, will use its values instead.

        :return: an initialized Contract instance
        """
        address = parse_address(address)
        proxy_config = Contract._create_proxy_config(proxy_config)

        abi = await ContractAbiResolver(
            address=address, client=client, proxy_config=proxy_config
        ).resolve()

        return Contract(address=address, abi=abi, client=client)

    @staticmethod
    async def declare(
        account: AccountClient,
        compiled_contract: str,
        *,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> DeclareResult:
        """
        Declares a contract

        :param account: An AccountClient used to sign and send declare transaction.
        :param compiled_contract: String containing compiled contract.
        :param max_fee: Max amount of Wei to be paid when executing transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :return: DeclareResult instance
        """
        declare_tx = await account.sign_declare_transaction(
            compiled_contract=compiled_contract,
            max_fee=max_fee,
            auto_estimate=auto_estimate,
        )
        res = await account.declare(transaction=declare_tx)

        return DeclareResult(
            hash=res.transaction_hash,
            _client=account.client,
            class_hash=res.class_hash,
            _account=account,
            compiled_contract=compiled_contract,
        )

    @staticmethod
    async def deploy_contract(
        account: AccountClient,
        class_hash: Hash,
        abi: List,
        constructor_args: Optional[Union[List, Dict]] = None,
        *,
        deployer_address: AddressRepresentation = DEFAULT_DEPLOYER_ADDRESS,
        max_fee: Optional[int] = None,
        auto_estimate: bool = False,
    ) -> "DeployResult":
        """
        Deploys a contract through Universal Deployer Contract

        :param account: An AccountClient used to sign and send deploy transaction.
        :param class_hash: The class_hash of the contract to be deployed.
        :param abi: An abi of the contract to be deployed.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param deployer_address: Address of the UDC. Is set to the address of
            the default UDC (same address on mainnet/testnet/devnet) by default.
            Must be set when using custom network other than ones listed above.
        :param max_fee: Max amount of Wei to be paid when executing transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :return: DeployResult instance
        """
        # pylint: disable=too-many-arguments
        deployer = Deployer(
            deployer_address=deployer_address, account_address=account.address
        )
        deploy_call, address = deployer.create_deployment_call(
            class_hash=class_hash, abi=abi, calldata=constructor_args
        )
        res = await account.execute(
            calls=deploy_call, max_fee=max_fee, auto_estimate=auto_estimate
        )

        deployed_contract = Contract(
            client=account,
            address=address,
            abi=abi,
        )
        deploy_result = DeployResult(
            hash=res.transaction_hash,
            _client=account.client,
            deployed_contract=deployed_contract,
        )

        return deploy_result

    @staticmethod
    async def deploy(
        client: Client,
        compilation_source: Optional[StarknetCompilationSource] = None,
        compiled_contract: Optional[str] = None,
        constructor_args: Optional[Union[List, Dict]] = None,
        salt: Optional[int] = None,
        search_paths: Optional[List[str]] = None,
    ) -> "DeployResult":
        # pylint: disable=too-many-arguments
        """
        Deploys a contract and waits until it has ``PENDING`` status.
        Either `compilation_source` or `compiled_contract` is required.

        :param client: Client
        :param compilation_source: string containing source code or a list of source files paths
        :param compiled_contract: string containing compiled contract. Useful for reading compiled contract from a file.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param salt: Optional salt. Random value is selected if it is not provided.
        :param search_paths: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts.
        :raises: `ValueError` if neither compilation_source nor compiled_contract is provided.
        :return: DeployResult instance

        .. deprecated:: 0.8.0
            This method has been deprecated in favor of deploying through cairo syscall.
            To deploy a contract use `Contract.deploy_contract`.
        """
        warnings.warn(
            "In the future versions of StarkNet, Deploy transaction will not be supported."
            "To deploy a contract use cairo syscall",
            category=DeprecationWarning,
        )

        compiled = create_compiled_contract(
            compilation_source, compiled_contract, search_paths
        )
        translated_args = translate_constructor_args(compiled.abi, constructor_args)
        deploy_tx = make_deploy_tx(
            compiled_contract=compiled,
            constructor_calldata=translated_args,
            salt=salt,
        )
        res = await client.deploy(deploy_tx)
        contract_address = res.contract_address

        deployed_contract = Contract(
            client=client,
            address=contract_address,
            abi=compiled.abi,
        )
        deploy_result = DeployResult(
            hash=res.transaction_hash,
            _client=client,
            deployed_contract=deployed_contract,
        )

        return deploy_result

    @staticmethod
    def compute_address(
        salt: int,
        compilation_source: Optional[StarknetCompilationSource] = None,
        compiled_contract: Optional[str] = None,
        constructor_args: Optional[Union[List, Dict]] = None,
        search_paths: Optional[List[str]] = None,
        deployer_address: int = 0,
    ) -> int:
        """
        Computes address for given contract.
        Either `compilation_source` or `compiled_contract` is required.

        :param salt: int
        :param compilation_source: string containing source code or a list of source files paths
        :param compiled_contract: string containing compiled contract. Useful for reading compiled contract from a file.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param search_paths: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts.
        :param deployer_address: address of the deployer (if not provided default 0 is used)

        :raises: `ValueError` if neither compilation_source nor compiled_contract is provided.

        :return: contract's address
        """
        # pylint: disable=too-many-arguments
        compiled = create_compiled_contract(
            compilation_source, compiled_contract, search_paths
        )
        translated_args = translate_constructor_args(compiled.abi, constructor_args)
        return compute_address(
            salt=salt,
            class_hash=compute_class_hash(compiled, hash_func=pedersen_hash),
            constructor_calldata=translated_args,
            deployer_address=deployer_address,
        )

    @staticmethod
    def compute_contract_hash(
        compilation_source: Optional[StarknetCompilationSource] = None,
        compiled_contract: Optional[str] = None,
        search_paths: Optional[List[str]] = None,
    ) -> int:
        """
        Computes hash for given contract.
        Either `compilation_source` or `compiled_contract` is required.

        :param compilation_source: string containing source code or a list of source files paths
        :param compiled_contract: string containing compiled contract. Useful for reading compiled contract from a file.
        :param search_paths: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts.
        :raises: `ValueError` if neither compilation_source nor compiled_contract is provided.
        :return:
        """
        compiled_contract = create_compiled_contract(
            compilation_source, compiled_contract, search_paths
        )
        return compute_class_hash(compiled_contract, hash_func=pedersen_hash)

    @classmethod
    def _make_functions(
        cls, contract_data: ContractData, client: Client
    ) -> FunctionsRepository:
        repository = {}

        for abi_entry in contract_data.abi:
            if abi_entry["type"] != "function":
                continue

            name = abi_entry["name"]
            repository[name] = ContractFunction(
                name=name,
                abi=abi_entry,
                contract_data=contract_data,
                client=client,
            )

        return repository

    @staticmethod
    def _create_proxy_config(proxy_config) -> ProxyConfig:
        if proxy_config is False:
            return ProxyConfig()
        proxy_arg = ProxyConfig() if proxy_config is True else proxy_config
        return prepare_proxy_config(proxy_arg)



from abc import ABC, abstractmethod
from typing import List

from starknet_py.net.models.transaction import Transaction
from starknet_py.net.models.typed_data import TypedData


class BaseSigner(ABC):
    """
    Base class for transaction signer. Implement methods from this ABC to use a custom signer in AccountClient
    """

    @property
    @abstractmethod
    def public_key(self) -> int:
        """
        Public key of the signer

        :return: public key
        """

    @abstractmethod
    def sign_transaction(self, transaction: Transaction) -> List[int]:
        """
        Sign execute transaction and return a signature

        :param transaction: Execute transaction to sign
        :return: transaction signature
        """

    @abstractmethod
    def sign_message(self, typed_data: TypedData, account_address: int) -> List[int]:
        """
        Sign an TypedData TypedDict for off-chain usage with the starknet private key and return the signature
        This adds a message prefix, so it can't be interchanged with transactions

        :param typed_data: TypedData TypedDict to be signed
        :param account_address: account address
        :return: the signature of the JSON object
        """



import warnings
from typing import Union, List, Optional

from starkware.starknet.definitions.fields import ContractAddressSalt

from starknet_py.common import create_compiled_contract
from starknet_py.compile.compiler import StarknetCompilationSource
from starknet_py.net.client_models import ContractClass
from starknet_py.net.models import Deploy


# pylint: disable=too-many-arguments
def make_deploy_tx(
    compilation_source: Optional[StarknetCompilationSource] = None,
    compiled_contract: Optional[Union[str, ContractClass]] = None,
    constructor_calldata: Optional[List[int]] = None,
    salt: Optional[int] = None,
    version: int = 0,
    cairo_path: Optional[List[str]] = None,
) -> Deploy:
    """
    Create deploy tx.
    Either `compilation_source` or `compiled_contract` is required.

    :param compilation_source: string containing source code or a list of source files paths
    :param compiled_contract: string containing compiled contract bytecode.
                              Useful for reading compiled contract from a file
    :param constructor_calldata: Calldata to be passed to contract constructor
    :param salt: Salt to be used when generating contract address
    :param version: PreparedFunctionCall version
    :param cairo_path: a ``list`` of paths used by starknet_compile to resolve dependencies within contracts
    :return: A "Deploy" transaction object

    .. deprecated:: 0.5.0
        Deploy transactions will not be supported in the future versions of StarkNet. Consider transitioning
        to Declare transactions and deploying through cairo syscall.
    """
    warnings.warn(
        "Deploy transactions will not be supported in the future versions of StarkNet. "
        "Consider transitioning to Declare transactions and deploying through cairo syscall.",
        category=DeprecationWarning,
    )

    if not constructor_calldata:
        constructor_calldata = []

    if isinstance(compiled_contract, str):
        compiled_contract = ContractClass.loads(compiled_contract)

    if not compiled_contract:
        compiled_contract = create_compiled_contract(
            compilation_source, compiled_contract, cairo_path
        )

    return Deploy(
        contract_address_salt=ContractAddressSalt.get_random_value()
        if salt is None
        else salt,
        contract_definition=compiled_contract,
        constructor_calldata=constructor_calldata,
        version=version,
    )



import pytest

from starknet_py.net.client_models import Call
from starknet_py.net.client_utils import _invoke_tx_to_call


call = Call(to_addr=1, selector=2, calldata=[3])


@pytest.mark.parametrize(
    "_call, invoke_tx, expected", [(call, None, call), (None, call, call)]
)
def test_invoke_tx_to_call(_call, invoke_tx, expected):
    new_call = _invoke_tx_to_call(call=_call, invoke_tx=invoke_tx)
    assert new_call == expected


def test_invoke_tx_to_call_raises_on_both_provided():
    with pytest.raises(ValueError) as exinfo:
        _invoke_tx_to_call(call=call, invoke_tx=call)

    assert "invoke_tx and call are mutually exclusive" in str(exinfo.value)


def test_invoke_tx_to_call_warns_on_invoke_tx():
    with pytest.warns(DeprecationWarning) as warn:
        _invoke_tx_to_call(invoke_tx=call)

    assert "invoke_tx parameter is deprecated. Use call instead" in str(warn[0].message)



from __future__ import annotations

from typing import Optional, List, Union, cast, NamedTuple

from starkware.starknet.definitions.fields import ContractAddressSalt
from starkware.starknet.public.abi import get_selector_from_name

from starknet_py.common import int_from_hex
from starknet_py.constants import DEFAULT_DEPLOYER_ADDRESS
from starknet_py.net.client_models import Hash, Call
from starknet_py.net.models import AddressRepresentation, parse_address, compute_address
from starknet_py.utils.contructor_args_translator import translate_constructor_args
from starknet_py.utils.crypto.facade import pedersen_hash
from starknet_py.utils.data_transformer.universal_deployer_serializer import (
    universal_deployer_serializer,
    deploy_contract_abi,
)
from starknet_py.utils.sync import add_sync_methods


ContractDeployment = NamedTuple("ContractDeployment", [("udc", Call), ("address", int)])


@add_sync_methods
class Deployer:
    """
    Deployer used to deploy contracts through Universal Deployer Contract (UDC)
    """

    def __init__(
        self,
        *,
        deployer_address: AddressRepresentation = DEFAULT_DEPLOYER_ADDRESS,
        account_address: Optional[AddressRepresentation] = None,
    ):
        """
        :param deployer_address: Address of the UDC. Is set to the address of
            the default UDC (same address on real nets and devnet) by default.
            Must be set when using custom network other than devnet.
        :param account_address: Should be equal to the address of the account which will send the transaction.
            If passed, it will be used to modify the salt, otherwise, salt will not be affected.
        """

        self.deployer_address = parse_address(deployer_address)
        self.account_address = account_address
        self._unique = account_address is not None

    def create_deployment_call(
        self,
        class_hash: Hash,
        *,
        salt: Optional[int] = None,
        abi: Optional[List] = None,
        calldata: Optional[Union[List, dict]] = None,
    ) -> ContractDeployment:
        """
        Creates deployment call to the UDC contract

        :param class_hash: The class_hash of the contract to be deployed
        :param salt: The salt for a contract to be deployed. Random value is selected if it is not provided
        :param abi: ABI of the contract to be deployed
        :param calldata: Constructor args of the contract to be deployed
        :return: NamedTuple with call and address of the contract to be deployed
        """
        if not abi and calldata:
            raise ValueError("calldata was provided without an abi")

        raw_calldata = translate_constructor_args(
            abi=abi or [], constructor_args=calldata
        )

        return self.create_deployment_call_raw(
            class_hash=class_hash, salt=salt, raw_calldata=raw_calldata
        )

    def create_deployment_call_raw(
        self,
        class_hash: Hash,
        *,
        salt: Optional[int] = None,
        raw_calldata: Optional[List[int]] = None,
    ) -> ContractDeployment:
        """
        Creates deployment call to the UDC contract with plain Cairo calldata

        :param class_hash: The class_hash of the contract to be deployed
        :param salt: The salt for a contract to be deployed. Random value is selected if it is not provided
        :param raw_calldata: Plain Cairo constructor args of the contract to be deployed
        :return: NamedTuple with call and address of the contract to be deployed
        """
        salt = cast(int, salt or ContractAddressSalt.get_random_value())
        class_hash = int_from_hex(class_hash)

        calldata, _ = universal_deployer_serializer.from_python(
            value_types=deploy_contract_abi["inputs"],
            classHash=class_hash,
            salt=salt,
            unique=int(self._unique),
            calldata=raw_calldata or [],
        )

        call = Call(
            to_addr=self.deployer_address,
            selector=get_selector_from_name("deployContract"),
            calldata=calldata,
        )

        address = self._compute_address(salt, class_hash, raw_calldata or [])

        return ContractDeployment(udc=call, address=address)

    def _compute_address(
        self, salt: int, class_hash: int, constructor_calldata: List[int]
    ) -> int:
        deployer_address = self.deployer_address if self._unique else 0
        salt = (
            pedersen_hash(parse_address(self.account_address), salt)
            if self.account_address is not None
            else salt
        )
        return compute_address(
            class_hash=class_hash,
            constructor_calldata=constructor_calldata,
            salt=salt,
            deployer_address=deployer_address,
        )



# This is needed for importing fixtures from `fixtures` directory
pytest_plugins = [
    "starknet_py.tests.e2e.fixtures.event_loop",
    "starknet_py.tests.e2e.fixtures.clients",
    "starknet_py.tests.e2e.fixtures.account_clients",
    "starknet_py.tests.e2e.fixtures.contracts",
    "starknet_py.tests.e2e.fixtures.misc",
    "starknet_py.tests.e2e.fixtures.devnet",
    "starknet_py.tests.e2e.fixtures.constants",
    "starknet_py.tests.e2e.fixtures.proxy",
    "starknet_py.tests.e2e.client.fixtures.transactions",
    "starknet_py.tests.e2e.client.fixtures.prepare_network",
]



import re
from abc import ABC, abstractmethod
from typing import Optional, Callable

from starkware.starknet.public.abi import (
    get_storage_var_address,
    get_selector_from_name,
)

from starknet_py.constants import RPC_INVALID_MESSAGE_SELECTOR_ERROR
from starknet_py.net.client import Client
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import Call
from starknet_py.net.models import Address


class ProxyCheck(ABC):
    @abstractmethod
    async def implementation_address(
        self, address: Address, client: Client
    ) -> Optional[int]:
        """
        :return: Implementation address of contract being proxied by proxy contract at `address`
            given as a parameter or None if implementation does not exist.
        """

    @abstractmethod
    async def implementation_hash(
        self, address: Address, client: Client
    ) -> Optional[int]:
        """
        :return: Implementation class hash of contract being proxied by proxy contract at `address`
            given as a parameter or None if implementation does not exist.
        """


class ArgentProxyCheck(ProxyCheck):
    async def implementation_address(
        self, address: Address, client: Client
    ) -> Optional[int]:
        return await self.get_implementation(
            address=address,
            client=client,
            get_class_func=client.get_class_hash_at,
            regex_err_msg=r"(is not deployed)",
        )

    async def implementation_hash(
        self, address: Address, client: Client
    ) -> Optional[int]:
        return await self.get_implementation(
            address=address,
            client=client,
            get_class_func=client.get_class_by_hash,
            regex_err_msg=r"(is not declared)",
        )

    @staticmethod
    async def get_implementation(
        address: Address, client: Client, get_class_func: Callable, regex_err_msg: str
    ) -> Optional[int]:
        call = ArgentProxyCheck._get_implementation_call(address=address)
        err_msg = r"(Entry point 0x[0-9a-f]+ not found in contract)|" + regex_err_msg
        try:
            (implementation,) = await client.call_contract(call=call)
            await get_class_func(implementation)
        except ClientError as err:
            if (
                re.search(err_msg, err.message, re.IGNORECASE)
                or err.code == RPC_INVALID_MESSAGE_SELECTOR_ERROR
            ):
                return None
            raise err
        return implementation

    @staticmethod
    def _get_implementation_call(address: Address) -> Call:
        return Call(
            to_addr=address,
            selector=get_selector_from_name("get_implementation"),
            calldata=[],
        )


class OpenZeppelinProxyCheck(ProxyCheck):
    async def implementation_address(
        self, address: Address, client: Client
    ) -> Optional[int]:
        return await self._get_storage_at_or_none(
            address=address, client=client, key="Proxy_implementation_address"
        )

    async def implementation_hash(
        self, address: Address, client: Client
    ) -> Optional[int]:
        return await self._get_storage_at_or_none(
            address=address, client=client, key="Proxy_implementation_hash"
        )

    @staticmethod
    async def _get_storage_at_or_none(
        address: Address, client: Client, key: str
    ) -> Optional[int]:
        result = await client.get_storage_at(
            contract_address=address,
            key=get_storage_var_address(key),
            block_hash="latest",
        )
        return result or None



import os
from typing import Optional

from crypto_cpp_py.cpp_bindings import (
    cpp_hash,
    get_cpp_lib_file,
    ECSignature,
)
from starkware.cairo.common.hash_state import compute_hash_on_elements
from starkware.cairo.lang.vm.crypto import pedersen_hash as default_hash
from starkware.crypto.signature.signature import sign

from starknet_py.net.client_models import Call

# PREFIX_TRANSACTION = encoded 'StarkNet Transaction'
PREFIX_TRANSACTION = 476441609247967894954472788179128007176248455022


def hash_call_with(call: Call, hash_fun):
    return compute_hash_on_elements(
        [
            call.to_addr,
            call.selector,
            compute_hash_on_elements(
                call.calldata,
                hash_func=hash_fun,
            ),
        ]
    )


# Interface
def use_cpp_variant() -> bool:
    force_disable_ext = (
        os.getenv("DISABLE_CRYPTO_C_EXTENSION", "false").lower() == "true"
    )
    cpp_lib_file = get_cpp_lib_file()
    return not force_disable_ext and bool(cpp_lib_file)


def message_signature(msg_hash, priv_key, seed: Optional[int] = 32) -> ECSignature:
    return sign(msg_hash, priv_key, seed)


def pedersen_hash(left: int, right: int) -> int:
    if use_cpp_variant():
        return cpp_hash(left, right)
    return default_hash(left, right)
